#!/usr/bin/env python
from __future__ import with_statement
import os
import subprocess
import sys
import urllib, urllib2, urlparse
from optparse import OptionParser

import bzrlib
from bzrlib import errors
from bzrlib.workingtree import WorkingTree

def check_revno(tree_dir, dep):
    if dep['revno'] == 'tip':
        return True

    wt = None
    try:
        try:
            wt = WorkingTree.open_containing(tree_dir)[0]
            wt.lock_read()
        except (errors.NoWorkingTree, errors.NotLocalUrl):
            print >> sys.stderr, "Error: no working tree at %s" % tree_dir
            return False
        if dep['revno'][:4] == 'tag:':
            tag = dep['revno'][4:]
            revid = wt.branch.tags.lookup_tag(tag)
            dep_revno_t = wt.branch.revision_id_to_dotted_revno(revid)
            dep_revno = ".".join(str(n) for n in dep_revno_t)
        else:
            dep_revno = dep['revno']
        revid = wt.last_revision()
        try:
            revno_t = wt.branch.revision_id_to_dotted_revno(revid)
        except errors.NoSuchRevision:
            print >> sys.stderr, "Error: could not map rev id '%s' to revno" % revid
            return False
        revno = ".".join(str(n) for n in revno_t)
    finally:
        if wt is not None:
            wt.unlock()

    match = revno == dep_revno
    if not match:
        print >> sys.stderr, "Warning: revno mismatch in %s, expected: %r but was: %r" % (dep['name'], dep_revno, revno)
    return match


class ProgressSpinner(object):
    """Un spinner *muy* simple para mostrar actividad en la terminal"""

    def __init__(self, out=sys.stderr):
        """inicializa el spinner"""
        self.out = out
        self.running = True
        def spinner_generator():
            while self.running:
                for i in ('|', '/', '-', '\\', '|', '/', '-', '\\'):
                    yield i
        self.spinner = spinner_generator()

    def next(self, *args):
        self.out.write('\r%s' % self.spinner.next())

    def finished(self):
        self.running = False
        self.out.write('\r' + ' '*70 + '\r')


def download(url, destdir, filename=None):
    """Copy the contents of a file from a given URL
    to a local file and creates the destination dir if don't exists.
    """
    # check if the url it's ok
    try:
        urllib2.urlopen(url)
    except  urllib2.HTTPError, e:
        print >>sys.stderr, e, ': ', url
        return
    if not os.path.exists(destdir):
        os.makedirs(destdir)
    if filename is None:
        filename = os.path.basename(urlparse.urlparse(url).path)
    dest = os.path.join(destdir, filename)
    p = ProgressSpinner()
    filename, headers = urllib.urlretrieve(url, dest, reporthook=p.next)
    p.finished()


def resolve_vars(conf):
    for k, v in conf.iteritems():
        if isinstance(v, (str, unicode)):
            prev = v
            curr = prev % conf
            while curr != prev:
                prev = curr
                curr = prev % conf
            conf[k] = curr


def download_cacerts():
    print "Fetching cacerts.pem"
    sys.stdout.flush()
    download('http://curl.haxx.se/ca/cacert.pem', external_dir)


def fetch_packages():
    bzrlib.initialize()
    all = config.packages + config.plugins
    for dep in all:
        resolve_vars(dep)
        tarball_filename = config.get_tarfile(dep)
        tarball_path = os.path.join(external_dir, tarball_filename)
        if dep.get('revno', None) != 'tip' and os.path.exists(tarball_path):
            print >> sys.stderr, "%s is up to date" % dep['name']
            continue
        elif dep.get('url', None):
            print "Fetching %s %s" % (dep['name'], dep['version'])
            sys.stdout.flush()
            download(dep['url'], external_dir, tarball_filename)
        elif dep.get('branch', None):
            tree_dir = os.path.join(external_dir, dep['name'])
            if dep['revno'] == 'tip':
                revno = []
                nice_revno = 'tip'
            else:
                revno = ['-r', dep['revno']]
                if dep['revno'][:4] == 'tag:':
                    nice_revno = dep['revno'][4:]
                else:
                    nice_revno = 'r' + dep['revno']
            cmd = ['bzr']
            print "Fetching %s %s" % (dep['name'], nice_revno)
            if os.path.exists(tree_dir):
                cmd += ['pull', '-q', '-d', tree_dir]
                cmd += revno
                subprocess.call(cmd)
            else:
                cmd += ['branch', '-q']
                cmd += revno + [dep['branch'], tree_dir]
                subprocess.call(cmd)
            if check_revno(tree_dir, dep):
                subprocess.call(['bzr', 'export', tarball_path, tree_dir])


if __name__ == '__main__':
    parser = OptionParser(usage="%prog [-c CONFIG_FILE]") 
    parser.add_option('-c', '--config', dest='config', action='store', default='config')
    (options, args) = parser.parse_args()

    try:
        global config
        config = __import__(options.config, globals(), locals(), [], 0)
    except ImportError:
        print "ERROR: failed to import config file"
        sys.exit(1)

    external_dir = os.path.join(config.base, 'src')

    print "Downloading to %s" % external_dir
    try:
        fetch_packages()
        download_cacerts()
    except KeyboardInterrupt as e:
        print "Download aborted"
